package audio;

import java.io.*;
import javax.sound.sampled.*;
import javax.swing.event.*;

/********************************************************/
/**
 * An AudioPlayer knows how to access audio hardware to capture
 * and playback sound.  It also knows how to access the file system
 * to read and write audio files.  It can read and write audio
 * files in multiple formats and with multiple audio data encodings.
 */
public class AudioPlayer {
    protected boolean isCapturing = false;
    protected boolean isPlaying = false;
    protected static short nPlayers = 0;
	protected AudioBuffer audioBuffer = null;
    protected String audioDescription = null;
	protected EventListenerList listenerList = new EventListenerList();
    
    protected static final AudioFormat defaultFormat =
    			new AudioFormat(8000,  // sample rate
                                16,    // sampleSizeInBits
                                1,     // channels
                                true,  // signed
                                true); // big endian 
	
	/********************************************************/
	/**
	 * Record audio into the internal buffer.
	 * This runs in a separate thread.
	 */
	public void captureAudio(){
		Thread newThread = new Thread(){
			public void run(){
				captureAudioWorker();
			}
		};
		newThread.start();		
	}
	private void captureAudioWorker(){
		int bufferSize;
		byte buffer[];
		TargetDataLine line = null;
		String errorString = null;

		// As this class only maintains one array of recorded data,
		// multiple instances of capture cannot be occuring simultaneously.
		// Play and record share both share this same array, so it is also
		// not possible to play and record at the same time.
		if (isCapturing){
			fireErrorEvent(" Multiple simultaneous captures not allowed.");
			return;
		}
		if (isPlaying){
			fireErrorEvent(" Simultaneous play and record not allowed.");
			return;
		}		
				
		audioBuffer = new AudioBuffer(defaultFormat);

		// set buffer size to 10ms for low latency
		bufferSize = ((int)(0.01 * defaultFormat.getSampleRate())) * defaultFormat.getFrameSize();
		buffer = new byte[bufferSize];
			    
		// Notify listeners that capture is starting.
		fireStateChangedEvent(State.RECORDING);
				
		// Begin capturing.
		isCapturing = true;
		try {
		    DataLine.Info info = new DataLine.Info(
		   	       	TargetDataLine.class, defaultFormat);
		    line = (TargetDataLine) AudioSystem.getLine(info);
		    line.open(defaultFormat);
		    line.start();
			        
		    while (isCapturing) {
		        int count = line.read(buffer, 0, buffer.length);
		        if (count > 0) {
		            audioBuffer.write(buffer, 0, count);
		        }
		    }
		    audioDescription = "Captured audio";
			        
		} catch (LineUnavailableException e) {
		    errorString =" Can't get an input line in captureAudio: " + e.getMessage();
		} finally {
		  	if (line != null){
		        line.stop();
		        line.close();
		    }
		    isCapturing = false;

		   	// Notify listeners that capture has stopped.
		    fireStateChangedEvent(State.IDLE);

		    if (errorString != null){
		    	fireErrorEvent(errorString);
			}
	    }
	}

	/********************************************************/
	/**
	 * Read audio from a file to the internal buffer.  This runs
	 * in a separate thread.
	 */
	public void openAudioFile(File readFile){
		final File localReadFile = readFile;
		Thread newThread = new Thread(){
			public void run(){
				openAudioFileWorker(localReadFile);
			}
		};
		newThread.start();		
	}	 
    private void openAudioFileWorker(File readFile){
        if (isCapturing || isPlaying){
        	fireErrorEvent(" Please stop audio before opening new file.");
        	return;
        }

		fireStateChangedEvent(State.OPENING);
		try{
 		 	audioBuffer = new AudioBuffer(
 		 	                  AudioSystem.getAudioInputStream(readFile)
			              );
 		 	audioDescription = readFile.getName();

 		 	fireStateChangedEvent(State.IDLE);
        } catch (UnsupportedAudioFileException a){
 		 	fireStateChangedEvent(State.IDLE);
        	fireErrorEvent(" Cannot read " + readFile.getName());
        } catch (IOException e) {
 		 	fireStateChangedEvent(State.IDLE);
            fireErrorEvent(" Error reading " + readFile.getName());
        }
    }


	/********************************************************/
	/**
	 * Play audio in the internal buffer.
	 * This runs in a separate thread.
	 */
	public void playAudio(){
		Thread newThread = new Thread(){
			public void run(){
				playAudioWorker();
			}
		};
		newThread.start();		
	}
    private void playAudioWorker(){
        byte buffer[];
        int bufferSize;
        SourceDataLine line = null;
        String errorString = null;
		AudioBuffer playBuffer;
          	
		// As play and record share both share the same audio data array, 
		// it is not possible to play and record at the same time.
		if (isCapturing){
			fireErrorEvent(" Simultaneous play and record not allowed.");
			return;
		}

		if (audioBuffer == null){
            fireErrorEvent(" No audio to play.");
            return;
        }

		// The SourceDataLine requires linear data for playback through
		// the audio hardware.  Transcode a copy of the audioBuffer's data to 
		// linear.

		playBuffer = new AudioBuffer(audioBuffer);
		playBuffer.transcode(AudioFormat.Encoding.PCM_SIGNED);

		// Notify listeners that playing is starting.
		nPlayers++;
		if (nPlayers == 1){
			fireStateChangedEvent(State.PLAYING);
		}
		
		// Begin playing
		isPlaying = true;
        try {       
	        // set buffer size to 10ms for low latency
	        bufferSize = ((int)(0.01 * playBuffer.getSampleRate())) 
	        	* playBuffer.getFrameSize();
	        buffer = new byte[bufferSize];
        
            DataLine.Info info = new DataLine.Info(
                   SourceDataLine.class, playBuffer.getFormat());
            line = (SourceDataLine) AudioSystem.getLine(info);
            line.open(playBuffer.getFormat());
            line.start();

            while (isPlaying){
            	int count;
            	
            	count = playBuffer.read(buffer, 0, buffer.length);

            	if (count < 0){
            		break;
            	}else if (count > 0){
            		line.write(buffer, 0, count);
            	}
            }
        } catch (IllegalArgumentException e){
        	errorString = " Unplayable audio format \n" + e.getMessage();
        } catch (LineUnavailableException e) {
        	errorString = " Can't get an output line in playAudio: " + e.getMessage();
        } catch (IOException e) {
            errorString = " Audio output problem: " + e.getMessage();
        } finally {
        	if (line != null){
	            line.drain();
	            line.close();
        	}
        	nPlayers--;
        	if (nPlayers == 0){
	        	isPlaying = false;
	        	// Notify listeners that playing has stopped.
	        	fireStateChangedEvent(State.IDLE);
	        }
        	
        	if (errorString != null){
        		fireErrorEvent(errorString);
        	}
        }
    }


	/********************************************************/
	/**
	 * Save audio in the internal buffer to a file after converting to
	 * the specified encoding.  This runs in a separate thread.
	 */
	public void saveAudioFile(File writeFile, AudioFormat.Encoding writeEncoding){
		final File localWriteFile = writeFile;
		final AudioFormat.Encoding localWriteEncoding = writeEncoding;
		Thread newThread = new Thread(){
			public void run(){
				saveAudioFileWorker(localWriteFile, localWriteEncoding);
			}
		};
		newThread.start();		
	}	 
	private void saveAudioFileWorker(File writeFile, AudioFormat.Encoding writeEncoding){
		
        if (isCapturing || isPlaying){
        	fireErrorEvent(" Please stop audio before saving.");
        	return;
        }
        
		if (audioBuffer == null){
			fireErrorEvent(" Nothing to save.");
			return;
		}        	
	    
	    fireStateChangedEvent(State.SAVING);

	    try {
	    	audioBuffer.transcode(writeEncoding);
			
			// save
	        FileOutputStream fos = new FileOutputStream(writeFile);

			AudioFileFormat.Type type = AudioFileFormat.Type.AU;
			if (writeFile.getName().endsWith(".wav")){
				type = AudioFileFormat.Type.WAVE;
			}

	    	AudioSystem.write(
	    		              audioBuffer.toAudioInputStream(), 
	    		              type,
	    		              fos
	    		             );
	        fos.close();
	        audioDescription = writeFile.getName();
	        fireStateChangedEvent(State.IDLE);
	        
	    } catch (Exception e) {
	    	System.out.println(e);
	    	fireStateChangedEvent(State.IDLE);
	        fireErrorEvent(" No file created");
	    }
	}
    
	/********************************************************/
	/**
	 * Save audio in the internal buffer to a file.  This runs
	 * in a separate thread.
	 */
    public void saveAudioFile(File writeFile) throws AudioPlayerException{
		saveAudioFile(writeFile, null);
    }

	/********************************************************/
	/**
	 * Return a boolean flag indicating whether the buffer is
	 * empty or not.
	 */
	public boolean hasAudio(){
		if (audioBuffer == null){
			return false;
		}else{
			return true;
		}
	}
	
	/********************************************************/
	/**
	 * Return an AudioFormat object specifying the type of
	 * audio dealt with by this class.
	 */
    public AudioFormat getFormat() {
    	if (audioBuffer != null){
	        return audioBuffer.getFormat();
	    }else{
	    	return null;
	    }
    }

	/********************************************************/
	/**
	 * Return a String describing the audio currently stored
	 * in the internal buffer (file name or other source).
	 */
	public String getAudioDescription(){
		if (audioDescription != null){
			return audioDescription;
		}else{
			return "";
		}
	}

	/********************************************************/
	/**
	 * Stop recording or playing.
	 */
    public void stop() {
		if (isCapturing == true || isPlaying == true){
	        isCapturing = false;
	        isPlaying = false;
	        fireStateChangedEvent(State.STOPPING);
	    }
    }
    
	/********************************************************/
	/**
	 * Get linearized data for graph.
	 */  
    public short[] getGraphData(){
    	byte[] bytes;
    	short[] shorts;
    	
    	if (audioBuffer == null) return new short[0];
    	
    	AudioBuffer playBuffer = new AudioBuffer(audioBuffer);
		playBuffer.transcode(AudioFormat.Encoding.PCM_SIGNED);

		bytes = playBuffer.toByteArray();
		shorts = new short[bytes.length/2];
		if (playBuffer.getFormat().isBigEndian()){
			for (int i=0; i<shorts.length; i++){
				shorts[i] = (short)((short)(bytes[2*i])*256 + (short)(bytes[2*i+1]));
			}
		}else{
			for (int i=0; i<shorts.length; i++){
				shorts[i] = (short)((short)(bytes[2*i+1])*256 + (short)(bytes[2*i]));
			}	
		}
		return shorts;
    }
    
    
	/********************************************************/
	/**
	 * Inner class to provide typesafe enumerated audio states.
	 */
	public static class State{
		private String desc;
		
		public static final State IDLE      = new State("idle");
		public static final State STOPPING  = new State("stopping");
		public static final State RECORDING = new State("recording");
		public static final State PLAYING   = new State("playing");
		public static final State OPENING   = new State("opening");
		public static final State SAVING    = new State("saving");
	
		private State(String s){
			desc = s;
		}
		
		public String toString(){
			return desc;
		}
	}
	
	/********************************************************/
   	/** 
	 * Allows other classes to listen for AudioPlayer events
	 */
	public void addAudioPlayerListener(AudioPlayerListener o){
		listenerList.add(AudioPlayerListener.class, o);
		o.audioStateChanged(State.IDLE);
	}
	
	/********************************************************/
    /** 
	 * Allows other classes to stop listening to AudioPlayer events
	 */

	public void removeAudioPlayerListener(AudioPlayerListener o){
		listenerList.remove(AudioPlayerListener.class, o);
	}

	/********************************************************/
    /** 
	 * Internally used for notifying listeners of state change events.
	 */

	private void fireStateChangedEvent(State newState){
		Object listeners[] = listenerList.getListeners(AudioPlayerListener.class);
		
		for (int i=0; i< listeners.length; i++){
			((AudioPlayerListener)listeners[i]).audioStateChanged(newState);
		}
	}

	/********************************************************/
    /** 
	 * Internally used for notifying listeners of errors.
	 */

	private void fireErrorEvent(String s){
		Object listeners[] = listenerList.getListeners(AudioPlayerListener.class);
		
		for (int i=0; i< listeners.length; i++){
			((AudioPlayerListener)listeners[i]).audioErrorOccurred(s);
		}
	}
}