package gui;

import java.awt.*;
import javax.swing.*;

/********************************************************/
/**
 * Oscilloscope display.
 */
public class ScopePanel extends JPanel{
	private short dataSet[] = null;
	private int xscale = 1;
	
	/********************************************************/
	/**
	 * Change the horizontal magnification.
	 *
	 * Zoom is limited to the range 0 to 6. 
	 * Magnification ratio is 1:N, where N=2^(zoom).
	 */
	 public void setZoom(int zoom){
	 	if (zoom < 0 || zoom > 6) return;
	 	xscale = (int)Math.pow(2, zoom);

	 	repaint();
	 }
	
	/********************************************************/
	/**
	 * Tell the scope what data to graph.
	 */
	public void setDataSet(short dataSet[]){
		this.dataSet = dataSet;
		
		repaint();
	}
	
	/********************************************************/
	/**
	 * Override JComponent.paint().
	 */
    public void paint(Graphics g) {
    	// read size of current visible area (assume user resized)
		Rectangle vr = getVisibleRect();
		
		// set size of underlying panel based on signal length and
		// visible area height
		int fullwidth;
		if (dataSet == null || dataSet.length == 0){
			fullwidth=0;
		}else{
			fullwidth=dataSet.length/xscale;
		}
		Dimension size = new Dimension(fullwidth, vr.height);
		setPreferredSize(size);

		// whitewash background of signal display
		g.setColor(Color.WHITE);
		g.fillRect(0,0,size.width,size.height);

		//if no signal, return.
		if (dataSet == null || dataSet.length == 0) return;
		
		// draw a zero reference
		int ymid = size.height/2;
		g.setColor(Color.RED);
		g.drawLine(0, ymid, size.width, ymid);
		
		// draw signal
		g.setColor(Color.BLACK);
		
		int i = vr.x;
		int n = vr.x * xscale;
		while ((i < vr.x+vr.width) && (n+xscale < dataSet.length)){
			int ymin = ymid -(int)((float)ymid/Short.MAX_VALUE*dataSetMin(n, n+xscale));
			int ymax = ymid -(int)((float)ymid/Short.MAX_VALUE*dataSetMax(n, n+xscale));
			
			g.drawLine(i, ymin, i, ymax);		
		
			i+=1;
			n+=xscale;		
		}
		
		revalidate();
   	}	
   
	/********************************************************/
	/**
	 * Helper function to find the minimum value of the dataset on
	 * an interval.
	 */
 	private short dataSetMin(int n1, int n2){
 		short y = Short.MAX_VALUE;
 		for (int i = n1; i<=n2; i++){	
 			if (dataSet[i]<y) y = dataSet[i];
 		}
 		return y;
 	}  

	/********************************************************/
	/**
	 * Helper function to find the maximum value of the dataset on
	 * an interval.
	 */
 	private short dataSetMax(int n1, int n2){
 		short y = Short.MIN_VALUE;
 		for (int i = n1; i<=n2; i++){	
 			if (dataSet[i]>y) y = dataSet[i];
 		}
 		return y;
 	}  
}