/************************************************************
 * Glenn Josefiak                                           *
 * SW511                                                    *
 * Fairfield University                                     *
 * Fall 2002                                                *
 ************************************************************/


package dsp;

import javax.sound.sampled.*;

/************************************************************/
/**
 * Computes MIDI notes closest to the supplied signed 16-bit
 * linear big-endian PCM audio.
 * This is done using a correlation with a sine wave at the
 * fundamental frequency of each of the notes -- essentially
 * a DFT with non-orthogonal basis functions.
 */
public class PitchDetector{
	private float sampleRate;
	private float presence[];
	private float window[] = new float[0];

	// MIDI note number for the lowest note recognized.
	private static final int MIDI_D2 = 38;

	// Assumed singing vocal range.
	private static final float note[] = {
		73.416f, 77.782f, 82.407f, 87.307f,	92.499f, 97.999f, // D2-G2
		103.83f, 110.00f, 116.54f, 123.47f, 130.81f, 138.59f, // G2#-C3#
		146.83f, 155.56f, 164.81f, 174.61f, 185.00f, 196.00f, // D3-G3
		207.65f, 220.00f, 233.08f, 246.94f, 261.63f, 277.18f, // G3#-C4#
		293.66f, 311.13f, 329.63f, 349.23f, 369.99f, 392.00f, // D4-G4
		415.30f, 440.00f, 466.16f, 493.88f, 523.25f, 554.37f  // G4#-C4#
	};

	/********************************************************/
	/**
	 * Construct a new PitchDetector.
	 */
	public PitchDetector(){
		presence = new float[note.length];
	}

	/********************************************************/
	/**
	 * Tell the PitchDetector what sample rate the supplied
	 * buffers of audio will contain.
	 */
	public void setSampleRate(float rate){
		sampleRate = rate;
	}

	/********************************************************/
	/**
	 * Process the supplied buffer and return the number of the
	 * note that correlates most highly with the input signal.
	 */
	public int compute(short buffer[]){
		float loudestComponent = 0;
		int loudestIndex = 0;
		float ww = (float)(2*Math.PI/sampleRate);
		
		float[] wb = new float[buffer.length]; // windowed buffer
		
		// Create a new window if necessary
		if (window.length != buffer.length){
			createWindow(buffer.length);
		}
		
		// Window the input signal.
		for (int i=0; i<buffer.length; i++){
			wb[i] = buffer[i] * window[i];
		}
		
		// Correlate the signal with a phasor at the fundamental
		// frequency of each of the notes of interest.
		for (int i = 0; i<note.length; i++){

			// Perform the correlation.
			float sac = 0, sas = 0;
			float xx = (float)(ww*note[i]);	
			for (int j = 0; j<wb.length; j++){
				sac += wb[j]*(float)Math.cos(xx*j);
				sas += wb[j]*(float)Math.sin(xx*j);
			}
			
			// Compute mean square of correlation (power at the note
			// of interest).
			presence[i] = (sac*sac + sas*sas)/wb.length;
			
			// Keep track of note that had the maximum power.
			if (presence[i] > loudestComponent){
				loudestComponent = presence[i];
				loudestIndex = i;
			}
		}
		
		// Do some heuristic manipulation to reduce the likelihood
		// that a fifth, fourth, third, or second harmonic will be
		// mistaken for the actual note.
		
		if (loudestIndex >27 &&
			presence[loudestIndex-28] > 0.1*presence[loudestIndex]){
			loudestIndex-=28;
		}else if (loudestIndex >23 &&
			presence[loudestIndex-24] > 0.1*presence[loudestIndex]){
			loudestIndex-=24;
		}else if (loudestIndex >18 &&
			presence[loudestIndex-19] > 0.1*presence[loudestIndex]){
			loudestIndex-=19;
		}else if (loudestIndex >11 && 
			presence[loudestIndex-12] > 0.1*presence[loudestIndex]){
			loudestIndex-=12;	
		}
		
		return loudestIndex+MIDI_D2;	
	}

	/********************************************************/
	/**
	 * Create a Hamming window of the specified length.
	 * See: Steiglitz, "A Digital Signal Processing Primer", p. 206
	 */
	private void createWindow(int len){
		window = new float[len];
		
		for (int i=0; i< len; i++){
			window[i] = (float)(0.54 - 0.46*Math.cos(2*Math.PI*i/(len-1)));
		}
	}
	
	/********************************************************/
	/**
	 * Obtain power results from the PitchDetector.
	 */
	public float[] getCorrelations(){
		return presence;
	}
}