package audio;

import java.io.*;
import javax.sound.sampled.*;
import javax.swing.event.*;

public class AudioPlayer {
    protected boolean isCapturing = false;
    protected boolean isPlaying = false;
    protected AudioFormat format = defaultFormat;
    protected byte audioByteArray[];
    private String audioDescription;
	private EventListenerList listenerList = new EventListenerList();
    
    protected static final AudioFormat defaultFormat =
    			new AudioFormat(8000,  // sample rate
                                16,    // sampleSizeInBits
                                1,     // channels
                                true,  // signed
                                true); // big endian 
	
	/********************************************************/
	/**
	 * Play audio in the internal buffer.  This is a time-consuming
	 * operation.  The calling program is responsible for running this
	 * operation in a separate thread.
	 */
    public void playAudio() throws AudioPlayerException {
        byte buffer[];
        int bufferSize;
        byte tempBuffer[] = null;
        int tempBufferSize;
        AudioInputStream ais;
        SourceDataLine line = null;
        String errorString = null;
        boolean uLaw = false;
        AudioFormat localFormat;
        
    	
		// As play and record share both share the same audio data array, 
		// it is not possible to play and record at the same time.
		if (isCapturing){
			throw new AudioPlayerException(" Simultaneous play and record not allowed.");
		}

		if (audioByteArray == null){
            throw new AudioPlayerException(" No audio to play.");
        }

        InputStream is = new ByteArrayInputStream(audioByteArray);
        ais = new AudioInputStream(
                is,
                format,
                audioByteArray.length / format.getFrameSize()
        );

		fireStateChangedEvent(State.PLAYING);
		isPlaying = true;
        try {
        	// Set up for playback and play until done.
        	// In general, transcoding to linear PCM needs to happen here
        	// for all compressed audio formats.
        	// For now, until I figure out JMF, I am making a special case
        	// of u-law data and leaving other compressed formats as
        	// unplayable. - GRJ 10/21/02
        
        	localFormat = format;
        	if (format.getEncoding() == AudioFormat.Encoding.ULAW){
        		localFormat = new AudioFormat(
					format.getSampleRate(), // sampleRate
					16,		// bits per sample
					1,		// channels; should this be more general??
					true,	// signed
					true	// big endian
				);
				
				// set buffer size to 10ms for low latency
				tempBufferSize = ((int)(0.01 * localFormat.getSampleRate())) * format.getFrameSize();
				tempBuffer = new byte[tempBufferSize];
				
				uLaw = true;
        	}        

	        // set buffer size to 10ms for low latency
	        bufferSize = ((int)(0.01 * localFormat.getSampleRate())) * localFormat.getFrameSize();
	        buffer = new byte[bufferSize];
        
            DataLine.Info info = new DataLine.Info(
                   SourceDataLine.class, localFormat);
            line = (SourceDataLine) AudioSystem.getLine(info);
            line.open(localFormat);
            line.start();

            while (isPlaying){
            	int count;
            	
           		if (uLaw){
           			// do u-law decompression on a sample-by-sample basis.
           			int sample;
           			
					count = ais.read(tempBuffer, 0, tempBuffer.length);
					for (int i=0; i<count; i++){
						sample = Ulaw.ulawToLinear(tempBuffer[i]);
						buffer[2*i] = (byte)(sample >> 8);
						buffer[2*i+1] = (byte)(sample & 0xFF);
					}
					count *= 2;								            		
           		}else{
           			// for other format, just play the buffer as is.
	            	count = ais.read(buffer, 0, buffer.length);
           		}

            	if (count < 0){
            		isPlaying = false;
            	}else if (count > 0){
            		line.write(buffer, 0, count);
            	}
            }
        } catch (IllegalArgumentException e){
        	errorString = " Unplayable audio format \n" + e.getMessage();
        } catch (LineUnavailableException e) {
        	errorString = " Can't get an output line in playAudio: " + e.getMessage();
        } catch (IOException e) {
            errorString = " Audio output problem: " + e.getMessage();
        } finally {
        	if (line != null){
	            line.drain();
	            line.close();
        	}
        	isPlaying = false;
        	fireStateChangedEvent(State.IDLE);
        	if (errorString != null){
        		throw new AudioPlayerException(errorString);
        	}
        }
    }


	/********************************************************/
	/**
	 * Record audio into the internal buffer.  This is a time-consuming
	 * operation.  The calling program is responsible for running this
	 * operation in a separate thread.
	 */
	public void captureAudio() throws AudioPlayerException {
       	int bufferSize;
	    byte buffer[];
	    TargetDataLine line = null;
	    String errorString = null;

		// As this class only maintains one array of recorded data,
		// multiple instances of capture cannot be occuring simultaneously.
		// Play and record share both share this same array, so it is also
		// not possible to play and record at the same time.
		if (isCapturing){
			throw new AudioPlayerException(" Multiple simultaneous captures not allowed.");
		}
		if (isPlaying){
			throw new AudioPlayerException(" Simultaneous play and record not allowed.");
		}		
		
	    format = defaultFormat;

        // set buffer size to 10ms for low latency
        bufferSize = ((int)(0.01 * format.getSampleRate())) * format.getFrameSize();
	    buffer = new byte[bufferSize];
	    
	    ByteArrayOutputStream recordedByteStream = new ByteArrayOutputStream();
		fireStateChangedEvent(State.RECORDING);
	    isCapturing = true;
	    try {
	        DataLine.Info info = new DataLine.Info(
	        	       	TargetDataLine.class, format);
	        line = (TargetDataLine) AudioSystem.getLine(info);
	        line.open(format);
	        line.start();
	        
	        while (isCapturing) {
	            int count = line.read(buffer, 0, buffer.length);
	            if (count > 0) {
	                recordedByteStream.write(
	                    buffer, 0, count);
	            }
	        }
	        recordedByteStream.close();
	        audioByteArray = recordedByteStream.toByteArray();
	        audioDescription = "Captured audio";
	        
	    } catch (LineUnavailableException e) {
	        errorString =" Can't get an input line in captureAudio: " + e.getMessage();
	    } catch (IOException e) {
	        errorString =" Audio input problem: " + e.getMessage();
	    } finally {
	    	if (line != null){
		        line.stop();
		        line.close();
		    }
		    isCapturing = false;
		    fireStateChangedEvent(State.IDLE);
		    if (errorString != null){
		    	throw new AudioPlayerException(errorString);
		    }
	    }
	}
    
	/********************************************************/
	/**
	 * Save audio in the internal buffer to a file.  This is a time-consuming
	 * operation.  The calling program is responsible for running this
	 * operation in a separate thread.
	 */
    public void saveAudioFile(String writeFileName) throws AudioPlayerException{
        if (isCapturing || isPlaying){
        	throw new AudioPlayerException(" Please stop audio before saving.");
        }
        
		if (audioByteArray == null){
			throw new AudioPlayerException(" Nothing to save.");
		}        	
	    
	    fireStateChangedEvent(State.SAVING);
        AudioInputStream ais = new AudioInputStream(
                new ByteArrayInputStream(audioByteArray),
                format,
                audioByteArray.length / format.getFrameSize()
        );

	    try {
	        FileOutputStream fos = new FileOutputStream(writeFileName);

	    	AudioSystem.write(ais, AudioFileFormat.Type.AU, fos);

	        fos.close();
	        audioDescription = writeFileName;
	        fireStateChangedEvent(State.IDLE);
	        
	    } catch (IOException ioe) {
	    	fireStateChangedEvent(State.IDLE);
	        throw new AudioPlayerException(" No file created");
	    }
    }

	/********************************************************/
	/**
	 * Read audio from a file to the internal buffer.  This is a time-consuming
	 * operation.  The calling program is responsible for running this
	 * operation in a separate thread.
	 */
    public void openAudioFile(String readFileName) throws AudioPlayerException {
		File readFile = null;

        if (isCapturing || isPlaying){
        	throw new AudioPlayerException(" Please stop audio before opening new file.");
        }

		fireStateChangedEvent(State.OPENING);
		try{
			readFile = new File(readFileName);
			AudioInputStream ais = AudioSystem.getAudioInputStream(readFile);

			audioByteArray = new byte[ais.available()];
			ais.read(audioByteArray, 0, ais.available());
 		 	format = ais.getFormat();
 		 	
 		 	audioDescription = readFile.getName();
 		 	fireStateChangedEvent(State.IDLE);
        } catch (UnsupportedAudioFileException a){
 		 	fireStateChangedEvent(State.IDLE);
        	throw new AudioPlayerException(" Cannot read " + readFile.getName());
        } catch (IOException e) {
 		 	fireStateChangedEvent(State.IDLE);
            throw new AudioPlayerException(" Error reading " + readFile.getName());
        }
    }

	/********************************************************/
	/**
	 * Return an AudioFormat object specifying the type of
	 * audio dealt with by this class.
	 */
    public AudioFormat getFormat() {
        return format;
    }

	/********************************************************/
	/**
	 * Return a String describing the audio currently stored
	 * in the internal buffer (file name or other source).
	 */
	public String getAudioDescription(){
		return audioDescription;
	}

	/********************************************************/
	/**
	 * Stop recording or playing.
	 */
    public void stop() {
        isCapturing = false;
        isPlaying = false;
        fireStateChangedEvent(State.STOPPING);
    }
    
	/********************************************************/
	/**
	 * Inner class to provide typesafe enumerated audio states.
	 */
	public static class State{
		private String desc;
		
		public static final State IDLE      = new State("idle");
		public static final State STOPPING  = new State("stopping");
		public static final State RECORDING = new State("recording");
		public static final State PLAYING   = new State("playing");
		public static final State OPENING   = new State("opening");
		public static final State SAVING    = new State("saving");
	
		private State(String s){
			desc = s;
		}
		
		public String toString(){
			return desc;
		}
	}
	
	/********************************************************/
   	/** 
	 * Allows other classes to listen for AudioPlayer events
	 */
	public void addAudioPlayerListener(AudioPlayerListener o){
		listenerList.add(AudioPlayerListener.class, o);
		o.audioStateChanged(State.IDLE);
	}
	
	/********************************************************/
    /** 
	 * Allows other classes to stop listening to AudioPlayer events
	 */

	public void removeAudioPlayerListener(AudioPlayerListener o){
		listenerList.remove(AudioPlayerListener.class, o);
	}

	/********************************************************/
    /** 
	 * Internally used for notifying listeners of events
	 */

	protected void fireStateChangedEvent(State newState){
		Object listeners[] = listenerList.getListeners(AudioPlayerListener.class);
		
		for (int i=0; i< listeners.length; i++){
			((AudioPlayerListener)listeners[i]).audioStateChanged(newState);
		}
	}
}